%=========================================================================%
%========   Compressive deconvolution in US imaging using SDMM    ========%
%========           Code Author: Zhouye Chen                      ========%
%========          Version (Date): July 29th, 2016                ========%
%========            Email: zhouye.c@gmail.com                    ========%
%=========================================================================%
%------------------------      COPYRIGHT      ----------------------------%
% Copyright (2016): Zhouye Chen, Adrian Basarab, Denis Kouam\'{e}.
% Permission to use, copy, modify, and distribute this software for any
% purpose without fee is hereby granted, provided that this entire notice
% is included in all copies of any software and in all copies of the
% supporting documentation for such software.
% This software is being provided "as is", without any express or iplied
% warranty. Inparticular, the authors do not make any representation or
% warranty of any kind concerning the merchantability of this software or
% its fitness for any particular purpose.
%-------------------------------------------------------------------------%
%-----------------------      REFERENCE      -----------------------------%
% This set of MATLAB files contain an implementation of the algorithms
% described in the following paper:
%
% Zhouye Chen, Adrian Basarab and Denis Kouam\'{e}, 
% "Reconstruction of Enhanced Ultrasound Images from Compressed 
%  Measurements Using Simultaneous Direction Method of Multipliers," 
% IEEE Trans. Ultrasonics, Ferroelectrics, and Frequency Control. 
%                                                           Accepted, 2016.
% Zhouye Chen, Adrian Basarab and Denis Kouam\'{e}, 
% "Compressive Deconvolution in Medical Ultrasound Imaging," 
% IEEE Trans. Medical Imaging. 35.3(2016):728-737.
%-------------------------------------------------------------------------%

if ~exist('predef_params','var')
    clear all; close all; clc;
end

%% Include necessary paths
path(path,'../images');
path(path,'../common');
% path(path,'../YALL1_v1.4');
path(path,'../SRMeasurement');

% Run_Me_1st %% Yall1

%% Initialization Parameters  
if ~exist('save_res','var')
    save_res = 0;        % 0/1 = (Don't save/ Save results)
end
if ~exist('fig','var')
    fig = 1;        % 0/1 = (Don't fig/ Fig results)
end
if ~exist('SpBasis','var')
    SpBasis = 1;             % 0/1 = (Fourier/Wavelets)
end
if ~exist('SNR','var')
    SNR = 40;              % Noise SNR
end
if ~exist('img_name','var')
    img_name = 'GGD15';              % Image Name
end

%% ADMM parameters
mu = 1e-5;
alpha = 2e-1; 
p = 1;
tol = 5e-3;
ItMax = 1e5;

%% RF data Simulation
% Tissue reflectivity function
load GGD1_15
x = refl;
x = x/max(x(:));
x_orig = x;
imsize = size(x);
m = size(x,1);
n = size(x,2);
N = prod(imsize);

% PSF
load psf
h = psf;

% Blur
h_ex = PSF_exten(size(x),h);
H = fft2(h_ex);
x_blur = real(ifft2(H.*fft2(x)));

cRatios = 0.4;%0.2:0.2:0.8;    % CS Ratios to be tested
seeds   = 3;%:3:15;       % Random number generator seeds
ssimK = [0.01 0.03];
window = fspecial('gaussian', 11, 1.5);
ssimL=1;

i = 0; j = 0;
for cRatio = cRatios
   i = i + 1;
   j = 0;
   K = N*cRatio; % Number of measurements
   for seed = seeds
       j = j + 1;    
       % measurement matrix definition
       rand_type = 0;
       trans_mode = 'FFT';%'BDCT';%'FFT';%'PFFT';%
       [select_vect, rand_vect, blk_size] = RnS_vct(N,K,rand_type,trans_mode);
       if strcmp(trans_mode,'PFFT')| strcmp(trans_mode,'FFT')
            % Define Sampling Operator;
            Phi = @(z) fft1d_f(z, select_vect, rand_vect);
            % Define the transpose of the Sampling Operator;
            Phit = @(z) fft1d_t(z, N, select_vect, rand_vect);  
            % Block-Based Operators:
       elseif strcmp(trans_mode,'BDCT') | strcmp(trans_mode,'BWHT')          
            % Define Sampling Operator;    
            Phi = @(z) blk_f1d(z, select_vect, rand_vect, trans_mode, blk_size);
            % Define the transpose of the Sampling Operator;
            Phit = @(z) blk_t1d(z, N, select_vect, rand_vect,trans_mode, blk_size); 
       end 
       
       %% Compressive Sampling: measurements
       y_cs = Phi(x_blur(:));
       % add noise
       [y,sigma2] = addNoise(y_cs,SNR,seed+2 ); 
       
       % define A in CS: A = Phi*Psi^{-1}
       if(SpBasis) % Wavelet
           wav_name = 'Haar';  % Wavelet Name
           wav_level = 3;
           [C0, S0] = wavedec2(zeros(imsize),wav_level,wav_name);
           W = @(z) WavTrfm(z, imsize, wav_level, wav_name);
           Wt = @(z) InvWavTrfm(z, S0, imsize, wav_name);
           N = length(wavedec2(x_blur,wav_level,wav_name));
       else % Fourier
           W = @(z) Fourier(z, imsize);
           Wt = @(z) InvFour(z, imsize);
       end
       A = @ (z) Phi(Wt(z));
       At = @ (z) W(Phit(z));  
       nonorth = check_orth(A, At, y);
      
       %% Main Algorithm
       % Yall1 for the initialization of a
%        Ayall.times = @(z)A(z);
%        Ayall.trans = @(z)At(z);
%        opts.rho = 1e-5;
%        opts.tol = 1e-3;
%        [a,out] = yall1(Ayall,y,opts);
%        x_blur_est = real(reshape(Wt(a),imsize));
%        BPSNR(i,j) = calcPSNR(x_blur_est/max(x_blur_est(:)), x_blur/max(x_blur(:)), 1)
%        Bssim(i,j) = ssim_index(x_blur_est/max(x_blur_est(:)), x_blur/max(x_blur(:)), ssimK, window, ssimL)
% %         x_blur_est = x_blur_est/max(x_blur_est(:));
       
        % zeros for initialization
       t0 = clock;    
       [x_estA, a_estA, kA] = CsDecADMM(y, H, W, Wt, A, At, imsize, N, mu, alpha, 1e0, 1e-2, p, ItMax, tol);
       tA(i,j) = etime(clock, t0);
       t1 = clock;
       [x_estS, a_estS, kS] = CsDecSDMM(y, H, W, Wt, Phi, Phit, imsize, N, mu, alpha, 1e-3, p, nonorth, ItMax, tol);
       tS(i,j) = etime(clock, t1);
       PSNRA(i,j) = calcPSNR(x_estA, x_orig, 1);
       ssimA(i,j) = ssim_index(x_estA, x_orig, ssimK, window, ssimL);
       PSNRS(i,j) = calcPSNR(x_estS, x_orig, 1);
       ssimS(i,j) = ssim_index(x_estS, x_orig, ssimK, window, ssimL);
       fprintf('CsRatio=%.3g, RandTest=%d: PSNRA=%.4g, SSIMA=%.4g, tA = %d, PSNRS=%.4g, SSIMS=%.4g, tS = %d\n  ',...
           cRatio, seed, PSNRA(i,j), 100*ssimA(i,j), tA(i,j), PSNRS(i,j), 100*ssimS(i,j),tS(i,j));
       N = prod(imsize);
   end
end

%% Display
if (fig)
    figure,subplot(221),imagesc(log10(5e-2+abs(hilbert(x_orig))));colormap gray; title('Original image')
    subplot(222),imagesc(log10(5e-2+abs(hilbert(x_blur)))); title('Blurred image')
    subplot(223),imagesc(log10(5e-2+abs(hilbert(x_estA)))); title('Estimated image (ADMM)')
    subplot(224),imagesc(log10(5e-2+abs(hilbert(x_estS)))); title('Estimated image (SDMM)')
end

%% save
if (save_res)
    [pth,fname,fext] = fileparts(img_name);
    save(['./Results/' fname '_SNR' num2str(SNR)]); 
end
       
